# Copyright (c) 2022-2023, NVIDIA CORPORATION.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import torch

from ...typing import CuGraphAnyHgCsr
from ...binding.pylibcugraphops_ext.operators import (
    agg_hg_basis_fg_n2n_post_fwd as fg_n2n_post_fwd,
    agg_hg_basis_fg_n2n_post_bwd as fg_n2n_post_bwd,
    agg_hg_basis_mfg_n2n_post_fwd as mfg_n2n_post_fwd,
    agg_hg_basis_mfg_n2n_post_bwd as mfg_n2n_post_bwd,
)


def agg_hg_basis_n2n_post(
    feat: torch.Tensor,
    weights_comb: Optional[torch.Tensor],
    graph: CuGraphAnyHgCsr,
    concat_own: bool = False,
    norm_by_out_degree: bool = False,
) -> torch.Tensor:
    r"""PyTorch autograd function for node-to-node RGCN-like basis regularized
    aggregation, with features being transformed after (post) this aggregation.

    Parameters
    ----------
    feat : torch.Tensor
        The input node features.
        Shape: (n_in_nodes, in_dim).

    weights_comb : torch.Tensor | None
        The combination weights. If `None`, no weights are used and features of
        different edge types are aggregated into separate output dimensions.
        Shape: (n_edge_types, n_bases) if provided.

    graph : fg_csr_hg_int[32|64] | mfg_csr_hg_int[32|64]
        The graph used for the operation.

    concat_own : bool, default=False
        Concatenate output node embeddings in the aggregation.

    norm_by_out_degree : bool, default=False
        If set, output embeddings are normed by the degree of the output node
        per edge type.

    Returns
    -------
    output : torch.Tensor
        The aggregation output.
        Shape: (n_out_nodes, out_dim) if `concat_own` is `True`;
               (n_out_nodes, out_dim + in_dim) otherwise;
               where out_dim = in_dim * n_bases if `weights_comb` is provided,
               otherwise, out_dim = in_dim * n_edge_types.

    Notes
    -----
    Replace n_in_nodes and n_out_nodes with n_nodes when given a full graph.
    """
    use_mfg = hasattr(graph, "sample_size")
    options = (concat_own, norm_by_out_degree)
    if use_mfg:
        return _agg_hg_basis_mfg_n2n_post_autograd.apply(
            feat, weights_comb, graph, options
        )

    return _agg_hg_basis_fg_n2n_post_autograd.apply(feat, weights_comb, graph, options)


class _agg_hg_basis_fg_n2n_post_autograd(torch.autograd.Function):
    r"""Custom autograd function for node-to-node full-graph RGCN-like basis
    regularized aggregation, with features being transformed after (post)
    this aggregation."""

    @staticmethod
    def forward(ctx, feat, weights_comb, fg, options):
        feat = feat.contiguous()
        concat_own, norm_by_out_degree = options
        dim = feat.shape[-1]
        n_bases = fg.n_edge_types if weights_comb is None else weights_comb.shape[-1]
        out_dim = dim * (n_bases + (1 if concat_own else 0))

        agg_output = torch.zeros(
            fg.n_nodes, out_dim, dtype=feat.dtype, device=feat.device
        )

        stream = torch.cuda.current_stream().cuda_stream
        if weights_comb is None:
            fg_n2n_post_fwd(
                agg_output,
                feat.detach(),
                None,
                fg,
                concat_own,
                norm_by_out_degree,
                stream_id=stream,
            )
        else:
            weights_comb = weights_comb.contiguous()
            fg_n2n_post_fwd(
                agg_output,
                feat.detach(),
                weights_comb.detach(),
                fg,
                concat_own,
                norm_by_out_degree,
                stream_id=stream,
            )

        ctx.fg = fg
        ctx.concat_own = concat_own
        ctx.norm_by_out_degree = norm_by_out_degree
        ctx.save_for_backward(feat, weights_comb)

        return agg_output

    @staticmethod
    def backward(ctx, grad_output):
        grad_output = grad_output.contiguous()
        feat, weights_comb = ctx.saved_tensors
        grad_feat = torch.zeros_like(feat)

        stream = torch.cuda.current_stream().cuda_stream
        if weights_comb is None:
            grad_weights_comb = None
            fg_n2n_post_bwd(
                grad_feat,
                None,
                grad_output,
                feat.detach(),
                None,
                ctx.fg,
                ctx.concat_own,
                ctx.norm_by_out_degree,
                stream_id=stream,
            )
        else:
            grad_weights_comb = torch.zeros_like(weights_comb)
            fg_n2n_post_bwd(
                grad_feat,
                grad_weights_comb,
                grad_output,
                feat.detach(),
                weights_comb.detach(),
                ctx.fg,
                ctx.concat_own,
                ctx.norm_by_out_degree,
                stream_id=stream,
            )

        return grad_feat, grad_weights_comb, None, None


class _agg_hg_basis_mfg_n2n_post_autograd(torch.autograd.Function):
    r"""Custom autograd function for node-to-node message-flow-graph RGCN-like
    basis regularized aggregation, with features being transformed after (post)
    this aggregation."""

    @staticmethod
    def forward(ctx, feat, weights_comb, mfg, options):
        feat = feat.contiguous()
        concat_own, norm_by_out_degree = options
        dim = feat.shape[-1]
        n_bases = mfg.n_edge_types if weights_comb is None else weights_comb.shape[-1]
        out_dim = dim * (n_bases + (1 if concat_own else 0))

        agg_output = torch.zeros(
            mfg.n_out_nodes, out_dim, dtype=feat.dtype, device=feat.device
        )

        stream = torch.cuda.current_stream().cuda_stream
        if weights_comb is None:
            mfg_n2n_post_fwd(
                agg_output,
                feat.detach(),
                None,
                mfg,
                concat_own,
                norm_by_out_degree,
                stream_id=stream,
            )
        else:
            weights_comb = weights_comb.contiguous()
            mfg_n2n_post_fwd(
                agg_output,
                feat.detach(),
                weights_comb.detach(),
                mfg,
                concat_own,
                norm_by_out_degree,
                stream_id=stream,
            )

        ctx.mfg = mfg
        ctx.concat_own = concat_own
        ctx.norm_by_out_degree = norm_by_out_degree
        ctx.save_for_backward(feat, weights_comb)

        return agg_output

    @staticmethod
    def backward(ctx, grad_output):
        grad_output = grad_output.contiguous()
        feat, weights_comb = ctx.saved_tensors

        grad_feat = torch.zeros_like(feat)
        stream = torch.cuda.current_stream().cuda_stream
        if weights_comb is None:
            grad_weights_comb = None
            mfg_n2n_post_bwd(
                grad_feat,
                None,
                grad_output,
                feat.detach(),
                None,
                ctx.mfg,
                ctx.concat_own,
                ctx.norm_by_out_degree,
                stream_id=stream,
            )
        else:
            grad_weights_comb = torch.zeros_like(weights_comb)
            mfg_n2n_post_bwd(
                grad_feat,
                grad_weights_comb,
                grad_output,
                feat.detach(),
                weights_comb.detach(),
                ctx.mfg,
                ctx.concat_own,
                ctx.norm_by_out_degree,
                stream_id=stream,
            )

        return grad_feat, grad_weights_comb, None, None



# Copyright (c) 2023, NVIDIA CORPORATION.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union, List, Tuple

import torch


def maybe_size(tensor: torch.Tensor, dim: int = -1):
    return tensor.size(dim) if tensor is not None else 0


def maybe_empty_like(input: torch.Tensor, create_tensor: bool = True):
    # intended for usages where a gradient is only initialized
    # if we have the corresponding ctx.needs_input_grads
    return torch.empty_like(input) if create_tensor else None


def maybe_empty(
    size: Union[List[int], Tuple[int], torch.Size],
    device: torch.device,
    dtype: torch.dtype,
    create_tensor: bool = True,
):
    # intended for usages where a gradient is only initialized
    # if we have the corresponding ctx.needs_input_grads
    return torch.empty(size, device=device, dtype=dtype) if create_tensor else None



# Copyright (c) 2023, NVIDIA CORPORATION.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Optional
from .utils import maybe_size, maybe_empty
import torch

from ...binding.pylibcugraphops_ext.operators import (
    update_efeat_fg_e2e_sum_fwd as sum_efeat_e2e_fwd,
    update_efeat_fg_e2e_sum_bwd as sum_efeat_e2e_bwd,
    update_efeat_fg_e2e_concat_fwd as concat_efeat_e2e_fwd,
    update_efeat_fg_e2e_concat_bwd as concat_efeat_e2e_bwd,
)


def update_efeat_e2e(
    edge_feat: Optional[torch.Tensor],
    src_node_feat: Optional[torch.Tensor],
    dst_node_feat: Optional[torch.Tensor],
    graph: Any,
    mode: str = "concat",
) -> torch.Tensor:
    r"""PyTorch autograd function for creating new edge features (update_efeat)
    based on either concatenating or summing edge features and the features of
    the corresponding source and destination node of each edge in an edge-to-edge
    fashion (e2e)

    Parameters
    ----------
    edge_feat : torch.Tensor | None
        The input edge features - ignored if passed as None
        Shape: (n_edges, dim_edge).

    src_node_feat : torch.Tensor | None
        The input source node features - ignored if passed as None
        Shape: (n_in_nodes, dim_src)

    dst_node_feat : torch.Tensor | None
        The input destination node features - ignored if passed as None
        Shape: (n_out_nodes, dim_dst)

    graph : fg_csr_int[32|64] | mfg_csr_int[32|64]
        The graph used for the operation.

    mode : str, default="concat"
        mode of update, either "concat" or "sum"

    Returns
    -------
    output : torch.Tensor
        The output after either concatenation or summation.
        Shape for "concat": (n_edges, dim_edge + dim_src + dim_dst).
        Shape for "sum": (n_edges, dim_edge).
    """

    modes = ["sum", "concat"]

    use_mfg = hasattr(graph, "sample_size")
    if mode.lower() not in modes:
        raise ValueError(f"Supported operations: {modes}, " f"but got '{mode}'.")

    if use_mfg:
        raise NotImplementedError("MFG version not available.")
    return _fg_e2e_autograd.apply(
        edge_feat, src_node_feat, dst_node_feat, (graph, mode)
    )


class _fg_e2e_autograd(torch.autograd.Function):
    r"""Custom autograd function for an update of the edge features based on
    concatenating or summing edge features, and source node and destination
    node features"""

    @staticmethod
    def forward(ctx, edge_feat, src_node_feat, dst_node_feat, args):
        fg, mode = args
        ctx.fg = fg
        ctx.mode = mode

        dim_edge = maybe_size(edge_feat)
        dim_src = maybe_size(src_node_feat)
        dim_dst = maybe_size(dst_node_feat)

        if dim_edge > 0:
            edge_feat = edge_feat.contiguous()
            dtype = edge_feat.dtype
            device = edge_feat.device

        elif dim_src > 0:
            src_node_feat = src_node_feat.contiguous()
            dtype = src_node_feat.dtype
            device = src_node_feat.device

        elif dim_dst > 0:
            dst_node_feat = dst_node_feat.contiguous()
            dtype = dst_node_feat.dtype
            device = dst_node_feat.device

        else:
            raise ValueError("forward called with all the inputs being None")

        stream = torch.cuda.current_stream().cuda_stream

        if mode.lower() == "concat":
            output = torch.empty(
                fg.n_indices, dim_edge + dim_dst + dim_src, dtype=dtype, device=device
            )
            concat_efeat_e2e_fwd(
                output,
                edge_feat.detach(),
                src_node_feat.detach(),
                dst_node_feat.detach(),
                fg,
                stream,
            )
            ctx.dim_edge = dim_edge
            ctx.dim_src = dim_src
            ctx.dim_dst = dim_dst

        else:
            output = torch.empty(fg.n_indices, dim_edge, dtype=dtype, device=device)
            sum_efeat_e2e_fwd(
                output,
                edge_feat.detach(),
                src_node_feat.detach(),
                dst_node_feat.detach(),
                fg,
                stream,
            )
            ctx.dim_edge = dim_edge
            ctx.dim_src = dim_edge
            ctx.dim_dst = dim_edge

        return output

    @staticmethod
    def backward(ctx, grad_output):
        grad_output = grad_output.contiguous()
        needs_grad_edge, needs_grad_src, needs_grad_dst, _ = ctx.needs_input_grad

        grad_edge_feat = maybe_empty(
            (ctx.fg.n_indices, ctx.dim_edge),
            dtype=grad_output.dtype,
            device=grad_output.device,
            create_tensor=needs_grad_edge,
        )
        grad_src_feat = maybe_empty(
            (ctx.fg.n_nodes, ctx.dim_src),
            dtype=grad_output.dtype,
            device=grad_output.device,
            create_tensor=needs_grad_src,
        )
        grad_dst_feat = maybe_empty(
            (ctx.fg.n_nodes, ctx.dim_dst),
            dtype=grad_output.dtype,
            device=grad_output.device,
            create_tensor=needs_grad_dst,
        )

        stream = torch.cuda.current_stream().cuda_stream

        if ctx.mode.lower() == "concat":
            concat_efeat_e2e_bwd(
                grad_edge_feat,
                grad_src_feat,
                grad_dst_feat,
                grad_output.detach(),
                ctx.fg,
                ctx.dim_edge,
                ctx.dim_src,
                ctx.dim_dst,
                stream,
            )

        else:
            sum_efeat_e2e_bwd(
                grad_edge_feat,
                grad_src_feat,
                grad_dst_feat,
                grad_output.detach(),
                ctx.fg,
                ctx.dim_edge,
                stream,
            )

        return grad_edge_feat, grad_src_feat, grad_dst_feat, None



# Copyright (c) 2023, NVIDIA CORPORATION.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch

from ...typing import CuGraphAnyCsr
from ...binding.pylibcugraphops_ext import ActivationOp, mha_params
from ...binding.pylibcugraphops_ext.operators import (
    activation_params,
    mha_gat_v2_fg_n2n_fwd as fg_n2n_fwd,
    mha_gat_v2_fg_n2n_bwd as fg_n2n_bwd,
)

_activation_ops = [x for x in dir(ActivationOp) if not x.startswith("__")]


def mha_gat_v2_n2n(
    feat: torch.Tensor,
    attn_weights: torch.Tensor,
    graph: CuGraphAnyCsr,
    num_heads: int = 1,
    activation: str = "LeakyReLU",
    negative_slope: float = 0.2,
    add_own_node: bool = False,
    concat_heads: bool = True,
) -> torch.Tensor:
    r"""PyTorch autograd function for a multi-head attention layer (GAT-like)
    without using cudnn (mha_gat_v2) with an activation prior to the dot
    product but none afterwards in a node-to-node reduction (n2n).

    Parameters
    ----------
    feat : torch.Tensor
        The input node features. Each row consists of concatenated features from
        different heads after the linear transformation.
        Shape: (n_in_nodes, dim_in) where dim_in = dim_head * num_heads, with
                dim_head being the feature dimension per head.

    attn_weights : torch.Tensor
        The attention weights.
        Shape: (dim,).

    graph : fg_csr_int[32|64] | mfg_csr_int[32|64]
        The graph used in for the operation.

    num_heads : int, default=1
        Number of heads in multi-head attention.

    activation : str, default="LeakyReLU"
        The activation function used in the attention mechanism. Choose from
        'ELU', 'LeakyReLU', 'Linear', 'ReLU', 'Scalar', 'Sigmoid' and 'Tanh'.

    negative_slope : float, default=0.2
        LeakyReLU angle of negative slope. Only effective when `activation` is
        `"LeakyReLU"`.

    add_own_node : bool, default=False
        Add self-loop for each node if set.

    concat_heads : bool, default=True
        Aggregated embeddings from each head are concatenated if `True` or
        averaged if `False`.

    Returns
    -------
    output : torch.Tensor
        The aggregation output.
        Shape: (n_out_nodes, dim_in) if `concat_heads` is `True`;
               (n_out_nodes, dim_head) otherwise.

    Notes
    -----
    Replace n_in_nodes and n_out_nodes with n_nodes when given a full graph.
    """
    if activation not in _activation_ops:
        raise ValueError(
            f"Supported activation functions: {_activation_ops}, "
            f"but got '{activation}'."
        )
    act_params = activation_params(getattr(ActivationOp, activation), negative_slope)
    params = mha_params(num_heads, act_params, add_own_node, concat_heads)

    use_mfg = hasattr(graph, "sample_size")
    if use_mfg:
        raise NotImplementedError
    return _mha_gat_v2_fg_n2n_autograd.apply(feat, attn_weights, graph, params)


class _mha_gat_v2_fg_n2n_autograd(torch.autograd.Function):
    r"""Custom autograd function for a multi-head attention layer (GAT-like)
    without using cudnn (mha_gat_v2) with an activation prior to the dot
    product but none afterwards operating on full graphs (fg) in a node-to-node
    reduction (n2n)."""

    @staticmethod
    def forward(ctx, feat, attn_weights, fg, params):
        feat = feat.contiguous()
        attn_weights = attn_weights.contiguous()
        dim = feat.shape[-1]
        dim_out = dim if params.concat_heads else (dim // params.num_heads)
        output = torch.zeros(
            fg.n_nodes, dim_out, dtype=feat.dtype, device=feat.device
        )
        sm_scores = torch.zeros(
            2, params.num_heads, fg.n_indices, dtype=feat.dtype, device=feat.device
        )
        act_scores = torch.zeros(
            fg.n_indices, dim, dtype=feat.dtype, device=feat.device
        )

        stream = torch.cuda.current_stream().cuda_stream
        fg_n2n_fwd(
            output,
            sm_scores,
            act_scores,
            feat.detach(),
            attn_weights.detach(),
            fg,
            params,
            stream_id=stream,
        )

        ctx.fg = fg
        ctx.params = params
        ctx.save_for_backward(feat, attn_weights, sm_scores, act_scores)

        return output

    @staticmethod
    def backward(ctx, grad_output):
        grad_output = grad_output.contiguous()
        feat, attn_weights, sm_scores, act_scores = ctx.saved_tensors

        grad_feat = torch.zeros_like(feat)
        grad_attn_weights = torch.zeros_like(attn_weights)
        grad_sm_scores = torch.zeros_like(sm_scores)

        stream = torch.cuda.current_stream().cuda_stream
        fg_n2n_bwd(
            grad_feat,
            grad_attn_weights,
            grad_sm_scores,
            grad_output,
            feat.detach(),
            attn_weights.detach(),
            sm_scores,
            act_scores,
            ctx.fg,
            ctx.params,
            stream_id=stream,
        )

        return grad_feat, grad_attn_weights, None, None




# This file was generated by 'versioneer.py' (0.26) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2023-03-24T13:53:30-0400",
 "dirty": false,
 "error": null,
 "full-revisionid": "505ea813fbc9d71a80faf63b1f83928fcfa77e97",
 "version": "23.02.01"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)



# Copyright (c) 2023, NVIDIA CORPORATION.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch

from ...typing import CuGraphAnyCsr
from ...binding.pylibcugraphops_ext.operators import (
    AggOp,
    agg_concat_fg_n2n_fwd as fg_n2n_fwd,
    agg_concat_fg_n2n_bwd as fg_n2n_bwd,
    agg_concat_mfg_n2n_fwd as mfg_n2n_fwd,
    agg_concat_mfg_n2n_bwd as mfg_n2n_bwd,
)

_aggr_ops = ["max", "mean", "min", "sum"]


def agg_concat_n2n(
    feat: torch.Tensor, graph: CuGraphAnyCsr, aggr: str = "sum"
) -> torch.Tensor:
    r"""PyTorch autograd function for simple aggregation using node features in
    an node-to-node reduction (n2n) while concatenating the original features of
    output at the end (agg_concat).

    Parameters
    ----------
    feat : torch.Tensor
        The input node features.
        Shape: (n_in_nodes, dim_in).

    graph : fg_csr_int[32|64] | mfg_csr_int[32|64]
        The graph used for the operation.

    aggr : str, default="sum"
        Aggregation operation. Choose from 'max', 'mean', 'min', 'sum'.

    Returns
    -------
    output : torch.Tensor
        The aggregation output.
        Shape: (n_out_nodes, 2 * dim_in).

    Notes
    -----
    Replace n_in_nodes and n_out_nodes with n_nodes when given a full graph.
    """
    use_mfg = hasattr(graph, "sample_size")
    if aggr not in _aggr_ops:
        raise ValueError(
            f"Supported aggregation operations: {_aggr_ops}, " f"but got '{aggr}'."
        )
    aggr = getattr(AggOp, aggr.capitalize())

    if use_mfg:
        return _agg_concat_mfg_n2n_autograd.apply(feat, graph, aggr)
    return _agg_concat_fg_n2n_autograd.apply(feat, graph, aggr)


class _agg_concat_fg_n2n_autograd(torch.autograd.Function):
    r"""Custom autograd function for simple aggregation using node features in
    an node-to-node reduction (n2n) on a full graph (fg) while concatenating the
    original features of output at the end (agg_concat)."""

    @staticmethod
    def forward(ctx, feat, fg, aggr):
        feat = feat.contiguous()
        dim_in = feat.shape[-1]
        agg_output = torch.zeros(
            fg.n_nodes, 2 * dim_in, dtype=feat.dtype, device=feat.device
        )

        if aggr in (AggOp.Max, AggOp.Min):
            idx_type = fg.__class__.__name__.split("_")[-1]
            out_pos = torch.zeros(
                fg.n_nodes, dim_in, dtype=getattr(torch, idx_type), device=feat.device
            )
        else:
            out_pos = None

        stream = torch.cuda.current_stream().cuda_stream
        fg_n2n_fwd(agg_output, feat.detach(), fg, aggr, out_pos, stream_id=stream)

        ctx.fg = fg
        ctx.aggr = aggr
        ctx.save_for_backward(feat, out_pos)

        return agg_output

    @staticmethod
    def backward(ctx, grad_output):
        grad_output = grad_output.contiguous()
        feat, out_pos = ctx.saved_tensors
        grad_feat = torch.zeros_like(feat)

        stream = torch.cuda.current_stream().cuda_stream
        fg_n2n_bwd(grad_feat, grad_output, ctx.fg, ctx.aggr, out_pos, stream_id=stream)

        return grad_feat, None, None


class _agg_concat_mfg_n2n_autograd(torch.autograd.Function):
    r"""Custom autograd function for simple aggregation using node features in
    an node-to-node reduction (n2n) on a message-flow-graph (mfg) while
    concatenating the original features of output at the end (agg_concat)."""

    @staticmethod
    def forward(ctx, feat, mfg, aggr):
        feat = feat.contiguous()
        dim_in = feat.shape[-1]
        agg_output = torch.zeros(
            mfg.n_out_nodes, 2 * dim_in, dtype=feat.dtype, device=feat.device
        )

        if aggr in (AggOp.Max, AggOp.Min):
            idx_type = mfg.__class__.__name__.split("_")[-1]
            out_pos = torch.zeros(
                mfg.n_out_nodes,
                dim_in,
                dtype=getattr(torch, idx_type),
                device=feat.device,
            )
        else:
            out_pos = None

        stream = torch.cuda.current_stream().cuda_stream
        mfg_n2n_fwd(agg_output, feat.detach(), mfg, aggr, out_pos, stream_id=stream)

        ctx.mfg = mfg
        ctx.aggr = aggr
        ctx.save_for_backward(feat, out_pos)

        return agg_output

    @staticmethod
    def backward(ctx, grad_output):
        grad_output = grad_output.contiguous()
        feat, out_pos = ctx.saved_tensors
        grad_feat = torch.zeros_like(feat)

        stream = torch.cuda.current_stream().cuda_stream
        mfg_n2n_bwd(
            grad_feat, grad_output, ctx.mfg, ctx.aggr, out_pos, stream_id=stream
        )

        return grad_feat, None, None



# Copyright (c) 2023, NVIDIA CORPORATION.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union

from .binding.pylibcugraphops_ext import (
    fg_csr_int32,
    fg_csr_int64,
    mfg_csr_int32,
    mfg_csr_int64,
    fg_csr_hg_int32,
    fg_csr_hg_int64,
    mfg_csr_hg_int32,
    mfg_csr_hg_int64,
)

FgCsr = Union[fg_csr_int32, fg_csr_int64]
MfgCsr = Union[mfg_csr_int32, mfg_csr_int64]
AnyCsr = Union[FgCsr, MfgCsr]
CuGraphFgCsr = FgCsr
CuGraphMfgCsr = MfgCsr
CuGraphAnyCsr = AnyCsr

FgHgCsr = Union[fg_csr_hg_int32, fg_csr_hg_int64]
MfgHgCsr = Union[mfg_csr_hg_int32, mfg_csr_hg_int64]
AnyHgCsr = Union[FgHgCsr, MfgHgCsr]
CuGraphFgHgCsr = FgHgCsr
CuGraphMfgHgCsr = MfgHgCsr
CuGraphAnyHgCsr = AnyHgCsr



# Copyright (c) 2022-2023, NVIDIA CORPORATION.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .agg_concat import agg_concat_n2n
from .agg_hg_basis import agg_hg_basis_n2n_post
from .mha_gat import mha_gat_n2n
from .mha_gat_v2 import mha_gat_v2_n2n
from .update_efeat import update_efeat_e2e

__all__ = [
    "agg_concat_n2n",
    "agg_hg_basis_n2n_post",
    "mha_gat_n2n",
    "mha_gat_v2_n2n",
    "update_efeat_e2e",
]
